import collections
import pprint

# Two simple node classes, chosen to be analogous to the example of a
# zipper for records in Clojure:
# http://tech.puredanger.com/2010/10/22/zippers-with-records-in-clojure/.
# The first has a field with a sequence of children, the second has
# fields which directly contain children.
class ScalarFunction(collections.namedtuple('ScalarFunction', 'f exprs')):
    @property
    def children(self):
        return self.exprs
    def make_node(self, children):
        return ScalarFunction(self.f, children)

class Comparison(collections.namedtuple('ScalarFunction', 'op left right')):
    @property
    def children(self):
        return (self.left, self.right)
    def make_node(self, children):
        return Comparison(self.op, *children)


# This follows to some extent the Clojure zipper implementions, https://github.com/clojure/clojure/blob/master/src/clj/clojure/zip.clj and https://github.com/akhudek/fast-zip/blob/master/src/cljx/fast_zip/core.cljx.  Clojure, like
# Python, is dynamically-typed, and the Clojure implementation also
# has an optimization that gives better asymptotic performance when
# going up if a node hasn't been changed, see my question at
# https://stackoverflow.com/questions/32363820/why-does-the-clojure-zipper-implementation-use-different-types-and-data-structur
# .  The use of tuples in this prototype will give worse asymptotic
# performance than using linked lists because some operations require
# copying.  The real implementation might use collections.deques
# (which are unnecessarily doubly-linked and memory-intensive, but
# ought to have the same asymptotic performance), a custom linked-list
# implementation, or just ignore the performance differences.

# I only implemented some of the operations implemented in Clojure, in
# the real implementation it would of course be possible to implement
# the rest, as well as some of the traversal iterators in
# https://github.com/trivio/zipper/blob/master/zipper.py , if they
# would be useful.  This prototype returns None when an operation is
# invalid, it might be better to raise some sort of exception for some
# or all of those cases.  (The Clojure implementation only raises
# exceptions in some cases.)
Path = collections.namedtuple('Path', 'left right parent_nodes parent_path changed')

class Zipper(collections.namedtuple('Location', 'node path')):
    def __new__(cls, root, path=None):
        return super(Zipper, cls).__new__(cls, root, path)

    def left(self):
        if self.path:
            node, siblings = self.path.left[-1], self.path.left[:-1]
            path = self.path._replace(left=siblings,
                                      right=(node,) + self.path.right)
            return self._replace(node=node, path=path)

    def right(self):
        if self.path:
            node, siblings = self.path.right[0], self.path.right[1:]
            path = self.path._replace(left=siblings + (node,), right=siblings)
            return self._replace(node=node, path=path)

    def down(self):
        children = getattr(self.node, 'children', None)
        if children:
            path = Path(
                left=(),
                right=children[1:],
                parent_nodes=self.path.parent_nodes + (self.node,) if self.path else (self.node,),
                parent_path=self.path,
                changed=False)
            return self._replace(node=children[0], path=path)

    def up(self):
        if self.path:
            left, right, parent_nodes, parent_path, changed = self.path
            if parent_nodes:
                node = parent_nodes[-1]
                if changed:
                    return self._replace(
                        node=node.make_node(left + (self.current,) + right),
                        path=parent_path and parent_path._replace(changed=True))
                else:
                    return self._replace(node=node, path=parent_path)

    def replace(self, node):
        return self._replace(node=node, path=self.path._replace(changed=True))

    def edit(self, *args, **kws):
        return self._replace(node=self.node.make_node(*args, **kws),
                             path=self.path._replace(changed=True))


if __name__ == '__main__':
    class Seq(tuple):
        def __new__(cls, *args):
            return super(Seq, cls).__new__(cls, args)
        @property
        def children(self):
            return self
        def make_node(self, children):
            return Seq(*children)

    z1 = Zipper(Seq(Seq('a', '*', 'b'), '+', Seq('c', '*', 'd')))
    print(z1.down().right().right().node)
    print(z1.down().right().right().down().right().node)
    print(z1.down().right().right().down().right().path)
    print(z1.down().right().right().down().right().up().up().node)

    print(z1.down().right().right().down().right().replace('/'))
    print(z1.down().edit(('e', '-', 'f')).node)
    
    f1 = ScalarFunction('+', (2, 3))
    f2 = ScalarFunction('-', (6, 1))
    c1 = Comparison('=', f1, f2)

    z2 = Zipper(c1)

    print(z1.down())
    print(z1.down().down())
    print(z1.down().down().right())
